@@ -2,7 +2,7 @@ import asyncio
 import re
 from collections import namedtuple
 from pathlib import Path
-from typing import Dict, Optional, Set
+from typing import Callable, Dict, List, Optional, Set
 from urllib.parse import urlsplit, urlunsplit
 
 import appdirs
@@ -30,38 +30,49 @@ def _get_urls_from_file(path):
 
 
 async def _get_return_code(
-    url: str, client, timeout: float, max_num_redirects: int = 10
+    url: str,
+    client,
+    timeout: float,
+    follow_codes: List[int],
+    max_num_redirects: int = 10,
+    is_allowed: Optional[Callable] = None,
 ):
     k = 0
     seq = []
     while True:
-        if k >= max_num_redirects:
-            seq.append(Info(997, url))
+        if is_allowed is not None and not is_allowed(url):
+            seq.append(Info(None, url))
             break
 
+        # Pretend to be a browser <https://stackoverflow.com/a/31597823/353337>.
+        # If we don't do this, sometimes we'll get a 403 where browsers don't (e.g.,
+        # JSTOR).
+        headers = {
+            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.182 Safari/537.36"
+        }
+
         try:
-            r = await client.head(url, allow_redirects=False, timeout=timeout)
-        except httpx.ConnectTimeout:
+            r = await client.head(
+                url, allow_redirects=False, timeout=timeout, headers=headers
+            )
+        except httpx.TimeoutException:
             seq.append(Info(998, url))
             break
-        except (
-            httpx.RemoteProtocolError,
-            httpx.ReadTimeout,
-            httpx.LocalProtocolError,
-            httpx.ConnectError,
-            httpx.ReadError,
-            httpx.PoolTimeout,
-        ):
+        except httpx.HTTPError:
+            # https://www.python-httpx.org/exceptions/
             seq.append(Info(999, url))
             break
 
         seq.append(Info(r.status_code, url))
 
-        if "Location" not in r.headers:
+        if (
+            k >= max_num_redirects
+            or r.status_code not in follow_codes
+            or "Location" not in r.headers
+        ):
             break
 
         # Handle redirect
-        assert 300 <= r.status_code < 400
         loc = r.headers["Location"]
         url_split = urlsplit(url)
 
@@ -95,7 +106,12 @@ async def _get_return_code(
 
 
 async def _get_all_return_codes(
-    urls, timeout: float, max_connections: int, max_keepalive_connections: int
+    urls,
+    timeout: float,
+    max_connections: int,
+    max_keepalive_connections: int,
+    follow_codes: List[int],
+    is_allowed: Optional[Callable] = None,
 ):
     # return await asyncio.gather(*map(_get_return_code, urls))
     ret = []
@@ -104,7 +120,12 @@ async def _get_all_return_codes(
         max_connections=max_connections,
     )
     async with httpx.AsyncClient(limits=limits) as client:
-        tasks = map(lambda x: _get_return_code(x, client, timeout), urls)
+        tasks = map(
+            lambda x: _get_return_code(
+                x, client, timeout, follow_codes=follow_codes, is_allowed=is_allowed
+            ),
+            urls,
+        )
         for task in track(
             asyncio.as_completed(tasks), description="Checking...", total=len(urls)
         ):
@@ -113,19 +134,26 @@ async def _get_all_return_codes(
     return ret
 
 
-def find_urls(paths):
-    urls = []
-    for path in paths:
-        path = Path(path)
-        if path.is_dir():
-            for p in path.rglob("*"):
-                if p.is_file():
-                    urls += _get_urls_from_file(p)
-        elif path.is_file():
-            urls += _get_urls_from_file(path)
-        else:
-            raise ValueError(f"Could not find path {path}")
-    return set(urls)
+def find_non_hidden_files(root):
+    root = Path(root)
+    if root.is_file():
+        if not root.name.startswith("."):
+            yield str(root)
+    else:
+        for path in root.glob("*"):
+            if not path.name.startswith("."):
+                if path.is_file():
+                    yield str(path)
+                else:
+                    yield from find_non_hidden_files(path)
+
+
+def find_files(paths: List[str]):
+    return [filepath for path in paths for filepath in find_non_hidden_files(path)]
+
+
+def find_urls(files):
+    return set(url for f in files for url in _get_urls_from_file(f))
 
 
 def replace_in_string(content: str, replacements: Dict[str, str]):
@@ -163,51 +191,34 @@ def replace_in_file(p, redirects: Dict[str, str]):
             f.write(new_content)
 
 
-def filter_urls(
-    urls: Set[str], allow_set: Optional[Set[str]], ignore_set: Optional[Set[str]]
-):
-    if allow_set is None:
-        allow_set = set()
-    if ignore_set is None:
-        ignore_set = set()
-
+def read_config():
     # check if there is a config file with more allowed/ignored domains
     config_file = Path(appdirs.user_config_dir()) / "deadlink" / "config.toml"
     try:
         with open(config_file) as f:
             out = toml.load(f)
     except FileNotFoundError:
-        pass
-    else:
-        if "allow" in out:
-            allow_set = allow_set.union(set(out["allow"]))
-        if "ignore" in out:
-            ignore_set = ignore_set.union(set(out["ignore"]))
+        out = {}
 
-    # filter out non-allowed and ignored urls
-    allowed_urls = set()
-    ignored_urls = set()
-    for url in urls:
-        is_allowed = True
+    return out
 
-        if is_allowed:
-            for a in allow_set:
-                if re.search(a, url) is None:
-                    is_allowed = False
-                    break
 
-        if is_allowed:
-            for i in ignore_set:
-                if re.search(i, url) is not None:
-                    is_allowed = False
-                    break
+def is_allowed(item, allow_set: Set[str], ignore_set: Set[str]):
+    is_allowed = True
 
-        if is_allowed:
-            allowed_urls.add(url)
-        else:
-            ignored_urls.add(url)
+    if is_allowed:
+        for a in allow_set:
+            if re.search(a, item) is None:
+                is_allowed = False
+                break
 
-    return allowed_urls, ignored_urls
+    if is_allowed:
+        for i in ignore_set:
+            if re.search(i, item) is not None:
+                is_allowed = False
+                break
+
+    return is_allowed
 
 
 def categorize_urls(
@@ -215,30 +226,47 @@ def categorize_urls(
     timeout: float = 10.0,
     max_connections: int = 100,
     max_keepalive_connections: int = 10,
+    is_allowed: Optional[Callable] = None,
 ):
+    # only follow permanent redirects
+    follow_codes = [301]
     r = asyncio.run(
-        _get_all_return_codes(urls, timeout, max_connections, max_keepalive_connections)
+        _get_all_return_codes(
+            urls,
+            timeout,
+            max_connections,
+            max_keepalive_connections,
+            follow_codes,
+            is_allowed,
+        )
     )
 
     # sort results into dictionary
     d = {
         "OK": [],
-        "Successful redirects": [],
-        "Failing redirects": [],
+        "Successful permanent redirects": [],
+        "Failing permanent redirects": [],
+        "Non-permanent redirects": [],
         "Client errors": [],
         "Server errors": [],
         "Timeouts": [],
         "Other errors": [],
+        "Ignored": [],
     }
     for item in r:
         status_code = item[0].status_code
-        if 200 <= status_code < 300:
+        if status_code is None:
+            d["Ignored"].append(item)
+        elif 200 <= status_code < 300:
             d["OK"].append(item)
         elif 300 <= status_code < 400:
-            if 200 <= item[-1].status_code < 300:
-                d["Successful redirects"].append(item)
+            if status_code == 301:
+                if 200 <= item[-1].status_code < 400:
+                    d["Successful permanent redirects"].append(item)
+                else:
+                    d["Failing permanent redirects"].append(item)
             else:
-                d["Failing redirects"].append(item)
+                d["Non-permanent redirects"].append(item)
         elif 400 <= status_code < 500:
             d["Client errors"].append(item)
         elif 500 <= status_code < 600:
@@ -266,13 +294,22 @@ def print_to_screen(d):
         num = len(d[key])
         console.print(f"{key} ({num})", style="green")
 
+    key = "Non-permanent redirects"
+    if key in d and len(d[key]) > 0:
+        print()
+        num = len(d[key])
+        console.print(f"{key} ({num})", style="green")
+
     key = "Ignored"
     if key in d and len(d[key]) > 0:
         print()
         num = len(d[key])
         console.print(f"{key} ({num})", style="white")
 
-    keycol = [("Successful redirects", "yellow"), ("Failing redirects", "red")]
+    keycol = [
+        ("Successful permanent redirects", "yellow"),
+        ("Failing permanent redirects", "red"),
+    ]
     for key, base_color in keycol:
         if key not in d or len(d[key]) == 0:
             continue
@@ -282,14 +319,16 @@ def print_to_screen(d):
             for k, item in enumerate(seq):
                 if item.status_code < 300:
                     color = "green"
-                elif 300 <= item.status_code < 400:
+                elif (300 <= item.status_code < 400) or item.status_code is None:
                     color = "yellow"
                 else:
                     color = "red"
+
+                sc = "xxx" if item.status_code is None else item.status_code
                 if k == 0:
-                    console.print(f"   {item.status_code}:   {item.url}", style=color)
+                    console.print(f"   {sc}:   {item.url}", style=color)
                 else:
-                    console.print(f"   → {item.status_code}: {item.url}", style=color)
+                    console.print(f"   → {sc}: {item.url}", style=color)
 
     for key in ["Client errors", "Server errors", "Timeouts", "Other errors"]:
         if key not in d or len(d[key]) == 0:
